@@ -2,7 +2,6 @@
 from __future__ import print_function
 from .tabulate import tabulate as tabulate_
 import sys
-import os.path
 import pandas as pd
 import re
 
@@ -37,35 +36,99 @@ A list of available commands follows.
 
 COMMANDS = {}
 
+
+def _tsv(*args, **kwargs):
+    kwargs["sep"] = "\t"
+    return pd.read_csv(*args, **kwargs)
+
+
 # These are all lambdas because they lazy load, and some of these
 # readers are introduced in later pandas.
 READERS = {
-    "csv": lambda fname: pd.read_csv(fname),
-    "csv_with_sep": lambda fname, sep: pd.read_csv(fname, sep),
-    "fwf": lambda fname: pd.read_fwf(fname),
-    "json": lambda fname: pd.read_json(fname),
-    "html": lambda fname: pd.read_html(fname),
-    "clipboard": lambda: pd.read_clipboard(),
+    "csv": pd.read_csv,
+    "clipboard": pd.read_clipboard,
+    "fwf": pd.read_fwf,
+    "json": pd.read_json,
+    "html": pd.read_html,
     "excel": lambda fname, sheet: pd.read_excel(fname, sheet),
     "xls": lambda fname, sheet: pd.read_excel(fname, sheet),
     "odf": lambda fname, sheet: pd.read_excel(fname, sheet),
-    "hdf5": lambda fname: pd.read_hdf(fname),
-    "feather": lambda fname: pd.read_feather(fname),
-    "parquet": lambda fname: pd.read_parquet(fname),
-    "orc": lambda fname: pd.read_orc(fname),
-    "msgpack": lambda fname: pd.read_msgpack(fname),
-    "stata": lambda fname: pd.read_stata(fname),
-    "sas": lambda fname: pd.read_sas(fname),
-    "spss": lambda fname: pd.read_spss(fname),
-    "pickle": lambda fname: pd.read_pickle(fname),
-    "sql": lambda fname: pd.read_sql(fname),
-    "gbq": lambda fname: pd.read_gbq(fname),
-    "google": lambda fname: pd.read_gbq(fname),
-    "bigquery": lambda fname: pd.read_gbq(fname),
-    ### extras
-    "tsv": lambda fname: pd.read_csv(fname, sep="\t"),
+    "tsv": _tsv,
 }
 
+
+try:
+    READERS["hdf5"] = pd.read_hdf
+except AttributeError:
+    pass
+
+
+try:
+    READERS["feather"] = pd.read_feather
+except AttributeError:
+    pass
+
+
+try:
+    READERS["parquet"] = pd.read_parquet
+except AttributeError:
+    pass
+
+
+try:
+    READERS["orc"] = pd.read_orc
+except AttributeError:
+    pass
+
+
+try:
+    READERS["msgpack"] = pd.read_msgpack
+except AttributeError:
+    pass
+
+
+try:
+    READERS["stata"] = pd.read_stata
+except AttributeError:
+    pass
+
+
+try:
+    READERS["sas"] = pd.read_sas
+except AttributeError:
+    pass
+
+
+try:
+    READERS["spss"] = pd.read_spss
+except AttributeError:
+    pass
+
+
+try:
+    READERS["pickle"] = pd.read_pickle
+except AttributeError:
+    pass
+
+
+try:
+    READERS["gbq"] = pd.read_gbq
+except AttributeError:
+    pass
+
+
+try:
+    READERS["google"] = pd.read_gbq
+except AttributeError:
+    pass
+
+
+try:
+    READERS["bigquery"] = pd.read_gb
+except AttributeError:
+    pass
+
+
 WRITERS = {
     "csv": "to_csv",
     "fwf": "to_fwf",
@@ -83,15 +146,18 @@ WRITERS = {
     "sas": "to_sas",
     "spss": "to_spss",
     "pickle": "to_pickle",
-    "sql": "to_sql",
     "gbq": "to_gbq",
     "google": "to_gbq",
     "bigquery": "to_gbq",
-    ### extras
+    # extras
     "tsv": "to_csv",
 }
 
 
+FALSY = ("False", "false", "No", "no", "0", False, 0)
+TRUTHY = ("True", "true", "Yes", "yes", "1", True, 1)
+
+
 def register(fn, name=None):
     if name is None:
         name = fn.__name__
@@ -222,13 +288,26 @@ def pipeout(df, sep=",", index=False, *args, **kwargs):
             pass
 
 
-def pipein(ftype="csv", sep=None):
+def pipein(ftype="csv", **kwargs):
+    skiprows = kwargs.get("skiprows")
+    if skiprows is not None:
+        try:
+            skiprows = int(skiprows)
+            if skiprows < 0:
+                raise ValueError("Negative")
+        except ValueError:
+            exit("skiprows must be a non-negative int, not {}".format(skiprows))
+        kwargs["skiprows"] = skiprows
+
+    sep = kwargs.get("sep")
+    thousands = kwargs.get("thousands")
+
     try:
         if sep is not None:
             if ftype != "csv":
                 exit("Only csv mode handles separators")
-            return READERS["csv_with_sep"](sys.stdin, sep)
-        return READERS[ftype](sys.stdin)
+            return READERS[ftype](sys.stdin, **kwargs)
+        return READERS[ftype](sys.stdin, **kwargs)
     except pd.errors.EmptyDataError:
         return pd.DataFrame()
 
@@ -378,6 +457,37 @@ def crosstab(column):
     pipeout(df)
 
 
+@register
+def groupby(*columns, how="sum", as_index=True):
+    """Group by columns, then apply `how` function.
+
+    Usage: cat a.csv | ph groupby animal --how=mean
+           cat a.csv | ph groupby animal --how=mean --as_index=False
+    """
+    columns = list(columns)
+    if not columns:
+        exit("Needs at least one column to group by")
+    df = pipein()
+    for c in columns:
+        if c not in df.columns:
+            exit("Unknown column name {}".format(c))
+    if as_index in TRUTHY:
+        as_index = True
+    elif as_index in FALSY:
+        as_index = False
+    else:
+        exit("--as_index=True or False, not {}".format(as_index))
+
+    grouped = df.groupby(columns, as_index=as_index)
+    try:
+        fn = getattr(grouped, how)
+    except AttributeError:
+        exit("Unknown --how={}, should be sum, mean, ...".format(how))
+    retval = fn()
+
+    pipeout(retval)
+
+
 @register
 def monotonic(column, direction="+"):
     """Check if a certain column is monotonically increasing or decreasing.
@@ -415,9 +525,9 @@ def iplot(*args, **kwargs):
 
     """
     try:
-        import cufflinks as cf
+        import cufflinks  # noqa
         import plotly as py
-    except ImportError as err:
+    except ImportError:
         exit("iplot needs cufflinks, pip install ph[iplot]")
 
     df = pipein()
@@ -435,6 +545,10 @@ def plot(*args, **kwargs):
             ph plot --kind=bar
             ph plot --kind=scatter --x=col1 --y=col2
             ph plot --style=k--
+            ph plot --logx=True
+            ph plot --logy=True
+            ph plot --loglog=True
+
     """
     try:
         import matplotlib.pyplot as plt
@@ -445,11 +559,18 @@ def plot(*args, **kwargs):
     index = kwargs.get("index")
     if index is not None:
         df = df.set_index(index)
+    for log_ in ("logx", "logy", "loglog"):
+        if kwargs.get(log_) in TRUTHY:
+            kwargs[log_] = True
+
     df.plot(
         kind=kwargs.get("kind", "line"),  # default pandas plot is line
         style=kwargs.get("style"),
         x=kwargs.get("x"),
         y=kwargs.get("y"),
+        logx=kwargs.get("logx"),
+        logy=kwargs.get("logy"),
+        loglog=kwargs.get("loglog"),
     )
     plt.show()
     pipeout(df)
@@ -486,7 +607,7 @@ def normalize(col=None):
 
 
 @register
-def date(col=None, unit=None, origin="unix", errors="raise"):
+def date(col=None, unit=None, origin="unix", errors="raise", dayfirst=False):
     """Assemble datetime from multiple columns or from one column
 
     --unit can be D, s, us, ns (defaults to ns, ns from origin)
@@ -497,6 +618,7 @@ def date(col=None, unit=None, origin="unix", errors="raise"):
 
     Usage: cat a.csv | ph date x
            cat a.csv | ph date x --unit=s --origin="1984-05-17 09:30"
+           cat a.csv | ph date x --dayfirst=True
            cat a.csv | ph date  # if a.csv contains year, month, date
 
     """
@@ -504,6 +626,8 @@ def date(col=None, unit=None, origin="unix", errors="raise"):
     if errors not in DATE_ERRORS:
         exit("Errors must be one of {}, not {}.".format(DATE_ERRORS, errors))
 
+    dayfirst = dayfirst in TRUTHY
+
     df = pipein()
     try:
         if col is None:
@@ -511,7 +635,9 @@ def date(col=None, unit=None, origin="unix", errors="raise"):
         else:
             if col not in df.columns:
                 exit("No such column {}".format(col))
-            df[col] = pd.to_datetime(df[col], unit=unit, origin=origin, errors=errors)
+            df[col] = pd.to_datetime(
+                df[col], unit=unit, origin=origin, errors=errors, dayfirst=dayfirst
+            )
     except Exception as err:
         exit(err)
 
@@ -544,7 +670,7 @@ def info():
 
 
 @register
-def to(ftype, fname=None, sep=None):
+def to(ftype, fname=None, sep=None, index=False):
     """Export csv to given format (possibly csv).
 
     Supports csv, html, json, parquet, bigquery, tsv, etc. (see README for full
@@ -552,6 +678,7 @@ def to(ftype, fname=None, sep=None):
 
     Usage: cat a.csv | ph to html
            cat a.csv | ph to tsv
+           cat a.csv | ph to csv --index=True
            cat a.csv | ph to csv --sep=';'
            cat a.csv | ph to clipboard
            cat a.csv | ph to json
@@ -565,12 +692,13 @@ def to(ftype, fname=None, sep=None):
         if ftype in ("parquet", "xls", "xlsx", "ods", "pickle"):
             exit("{} needs a path".format(ftype))
 
-    if ftype == "sql":
-        exit("sql writer not implemented")
-
     if ftype == "hdf5":
         exit("hdf5 writer not implemented")
 
+    if index not in TRUTHY + FALSY:
+        exit("Index must be True or False, not {}".format(index))
+    index = index in TRUTHY
+
     if ftype == "fwf":
         # pandas has not yet implemented to_fwf
         df = pipein()
@@ -594,14 +722,18 @@ def to(ftype, fname=None, sep=None):
         kwargs["sep"] = "\t"
     elif ftype == "csv" and sep is not None:
         kwargs["sep"] = sep
+
+    if ftype == "json":
+        index = True
+
     if fname is not None:
-        print(fn(fname, **kwargs))
+        print(fn(fname, index=index, **kwargs))
     else:
-        print(fn(**kwargs))
+        print(fn(index=index, **kwargs))
 
 
 @registerx("from")
-def from_(ftype="csv", sep=None):
+def from_(ftype="csv", **kwargs):
     """Read a certain (default csv) format from standard in and stream out as csv.
 
     Usage: cat a.json | ph from json
@@ -610,18 +742,42 @@ def from_(ftype="csv", sep=None):
 
     cat a.csv
     cat a.csv | ph to json | ph from json
+    cat a.tsv | ph from tsv
+    cat a.tsv | ph from csv --sep='\t'
+    cat a.tsv | ph from csv --sep='\t' --thousands=','
+
 
     """
-    if ftype == "clipboard":
-        pipeout(READERS["clipboard"]())
-    else:
-        if sep is not None:
-            if ftype != "csv":
-                exit("Only csv mode accepts separator")
-            pipeout(pipein(ftype, sep=sep))
+    skiprows = kwargs.get("skiprows")
+    if skiprows is not None:
+        try:
+            skiprows = int(skiprows)
+            if skiprows < 0:
+                raise ValueError("Negative")
+        except ValueError:
+            exit("skiprows must be a non-negative int, not {}".format(skiprows))
+        kwargs["skiprows"] = skiprows
 
-        else:
-            pipeout(pipein(ftype))
+    sep = kwargs.get("sep")
+    thousands = kwargs.get("thousands")
+
+    if ftype == "clipboard":
+        pipeout(READERS["clipboard"], **kwargs)
+        return
+
+    if sep is not None:
+        if ftype != "csv":
+            exit("Only csv mode accepts separator")
+        pipeout(pipein(ftype, **kwargs))
+        return
+
+    if thousands is not None:
+        if ftype != "csv":
+            exit("Only csv mode accepts thousands")
+        pipeout(pipein(ftype, **kwargs))
+        return
+
+    pipeout(pipein(ftype, **kwargs))
 
 
 @register
@@ -738,24 +894,66 @@ def help(*args, **kwargs):
     try:
         fn = getattr(ph, cmd)
         ds = getattr(fn, "__doc__")
-    except AttributeError as err:
+    except AttributeError:
         try:
             fn = getattr(pd.DataFrame, cmd)
             ds = getattr(fn, "__doc__")
-        except AttributeError as err:
+        except AttributeError:
             ds = ""
 
     print("Usage: ph {} [?]".format(cmd))
     print("       {}".format(ds))
 
 
+def slugify_name(name):
+    if not name:
+        return "unnamed"
+    if name == "_":
+        return "_"
+    lead_under = name[0] == "_"
+    trail_under = name[-1] == "_"
+
+    name = name.strip().lower()
+    unwanted = set(c for c in name if not c.isalnum())
+    for u in unwanted:
+        name = name.replace(u, "_").strip()
+    while "__" in name:
+        name = name.replace("__", "_").strip()
+    name = name.strip("_")
+    if lead_under:
+        name = "_" + name
+    if trail_under:
+        name = name + "_"
+    return name
+
+
+@register
+def slugify():
+    """Slugify the column headers.
+
+    Usage: cat a.csv | ph slugify
+
+    Removes all non-alphanumeric characters aside from the underscore.
+
+    Is useful in scenarios where you have possibly many columns with
+    very ugly names.  Can be a good preprocessor to @rename:
+
+    Usage: cat a.csv | ph slugify | ph rename less_bad_name good_name
+
+    """
+    df = pipein()
+    df.columns = [slugify_name(name) for name in df.columns]
+    pipeout(df)
+
+
 @registerx("open")
-def open_(ftype, fname=None, sheet=0, sep=None):
+def open_(ftype, fname, **kwargs):
     """Use a reader to open a file.
 
     Open ftype file with name fname and stream out.
 
     Usage: ph open csv a.csv
+           ph open csv a.csv --skiprows=7
            ph open json a.json
            ph open parquet a.parquet
            ph open excel a.ods
@@ -763,24 +961,36 @@ def open_(ftype, fname=None, sheet=0, sep=None):
            ph open excel a.xlsx
            ph open excel a.xls --sheet=2
            ph open excel a.xls --sheet="The Real Dataset sheet"
+           ph open csv a.csv --thousands=','
 
 
     """
     if ftype not in READERS:
         exit("Unknown filetype {}".format(ftype))
     reader = READERS[ftype]
-    kwargs = {}
+
+    sheet = kwargs.get("sheet")
     if ftype in ("excel", "xls", "odf"):
         kwargs["sheet"] = __tryparse(sheet)
-    if ftype == "csv" and sep is not None:
-        reader = READERS["csv_with_sep"]
-        kwargs["sep"] = sep
+
+    sep = kwargs.get("sep")
+    thousands = kwargs.get("thousands")
 
     if ftype == "clipboard" and fname is not None:
         exit("clipboard does not take fname")
     if ftype != "clipboard" and fname is None:
         exit("filename is required for {}".format(ftype))
 
+    skiprows = kwargs.get("skiprows")
+    if skiprows is not None:
+        try:
+            skiprows = int(skiprows)
+            if skiprows < 0:
+                raise ValueError("Negative")
+        except ValueError:
+            exit("skiprows must be a non-negative int, not {}".format(skiprows))
+        kwargs["skiprows"] = skiprows
+
     try:
         if ftype == "clipboard":
             df = reader(**kwargs)
@@ -950,7 +1160,7 @@ def sort(col):
 
     """
     df = pipein()
-    if not col in df.columns:
+    if col not in df.columns:
         exit("Unknown column {}".format(col))
     pipeout(df.sort_values(col))
 
